# Copyright (c) 2012-2017 Snowflake Computing Inc. All rights reserved.
"""
simple_ingest_manager - Creates a basic ingest manager that synchronously sends
requests to the Snowflake Ingest service
"""

# This class will manage our tokens
from .utils import SecurityManager

# This will manage our URL generation
from .utils import URLGenerator
from .utils.network import SnowflakeRestful
from .utils.uris import DEFAULT_HOST_FMT
from .utils.uris import DEFAULT_PORT
from .utils.uris import DEFAULT_SCHEME
from .version import __version__

# We use a named tuple to represent remote files
from collections import namedtuple

# Typing information
from typing import Text, Dict, Any

# UUID for typing formation
from uuid import UUID

import sys
import platform

from logging import getLogger
logger = getLogger(__name__)

# We just need a simple named tuple to represent remote files
StagedFile = namedtuple("StagedFile", ["path", "size"])

AUTH_HEADER = "Authorization"  # Authorization header name
BEARER_FORMAT = "BEARER {0}"  # The format of this bearer

USER_AGENT_HEADER = "User-Agent" # User-Agent header name
CLIENT_NAME = u"SnowpipePythonSDK" # Don't change!
CLIENT_VERSION = __version__
PLATFORM = platform.platform()
PYTHON_VERSION = u'.'.join(str(v) for v in sys.version_info[:3])
SNOWPIPE_SDK_USER_AGENT = \
        u'{name}/{version}/{python_version}/{platform}'.format(
                name=CLIENT_NAME,
                version=CLIENT_VERSION,
                python_version=PYTHON_VERSION,
                platform=PLATFORM)

OK = 200  # Is this Response OK?


class SimpleIngestManager(object):
    """
    SimpleIngestManager - this class is a simple wrapper around the Snowflake Ingest
    Service rest api. It is *synchronous* and as such we will block until we either totally fail to
    get a response *or* we successfully hear back from the
    """

    def __init__(self, account: Text, user: Text, pipe: Text, private_key: Text,
                 scheme: Text = DEFAULT_SCHEME, host: Text = None, port: int = DEFAULT_PORT):
        """
        Simply instantiates all of our local state
        :param account: the name of the account who is loading
        :param user: the name of the user who is loading
        :param pipe: the name of the pipe which we want to use for ingesting
        :param private_key: the private key we use for token signature
        """
        self.sec_manager = SecurityManager(account, user, private_key)  # Create the token generator
        self.url_engine = URLGenerator(scheme=scheme,
                                       host=host if host is not None else DEFAULT_HOST_FMT.format(account),
                                       port=port)
        self.pipe = pipe
        self._next_begin_mark = None
        self.restful = SnowflakeRestful()

    def _get_auth_header(self) -> Dict[Text, Text]:
        """
        _get_auth_header - simply method to generate the bearer header for our http requests
        :return: A singleton mapping from bearer to token
        """

        token_bearer = BEARER_FORMAT.format(self.sec_manager.get_token())
        return {AUTH_HEADER: token_bearer}

    def _get_user_agent_header(self) -> Dict[Text, Text]:
        """
        _get_user_agent_header - method to generate the user-agent header for our http requests
        :return: A singleton mapping for user agent and sdk version
        """
        return {USER_AGENT_HEADER: SNOWPIPE_SDK_USER_AGENT}

    def _get_headers(self) -> Dict[Text, Text]:
        """
        _get_headers - get all required SDK headers to be sent to the service
        :return: Array of headers to be sent
        """
        headers = self._get_auth_header()
        headers.update(self._get_user_agent_header())
        return headers

    def ingest_files(self, staged_files: [StagedFile], request_id: UUID = None) -> Dict[Text, Any]:
        """
        ingest_files - Informs Snowflake about the files to be ingested into a table through this pipe
        :param staged_files: a list of files we want to ingest
        :param request_id: an optional request uuid to label this request
        :return: the deserialized response from the service
        """
        # Generate the target url
        target_url = self.url_engine.make_ingest_url(self.pipe, request_id)
        logger.info('Ingest file request url: %s', target_url)

        # Make our message payload
        payload = {
            "files": [x._asdict() for x in staged_files]
        }

        # Send our request!
        headers = self._get_headers()
        response_body = self.restful.post(target_url, json=payload, headers=headers)
        logger.debug('Ingest response: %s', str(response_body))

        return response_body

    def get_history(self, recent_seconds: int = None, request_id: UUID = None) -> Dict[Text, Any]:
        """
        get_history - returns the currently cached ingest history from the service
        :param request_id: an optional request UUID to label this
        :param recent_seconds: an optional argument that specify the time range that history can be seen
        :return: the deserialized response from the service
        """
        # generate our history endpoint url
        target_url = self.url_engine.make_history_url(self.pipe, recent_seconds, self._next_begin_mark, request_id)
        logger.info('Get history request url: %s', target_url)

        # Send out our request!
        headers = self._get_headers()
        response_body = self.restful.get(target_url, headers=headers)

        self._next_begin_mark = response_body['nextBeginMark']
        
        return response_body

    def get_history_range(self, start_time_inclusive: Text, end_time_exclusive: Text = None,
            request_id: UUID = None) -> Dict[Text, Any]:
        """
        get_history_range - returns the ingest history between two points in time
        :param request_id: an optional request UUID to label this
        :param start_time_inclusive: Timestamp in ISO-8601 format. Start of the time range to retrieve load history data.
        :param end_time_exclusive: Timestamp in ISO-8601 format. End of the time range to retrieve load history data.
                                    If omitted, then CURRENT_TIMESTAMP() is used as the end of the range.
        :return: the deserialized response from the service
        """
        # generate our history endpoint url
        target_url = self.url_engine.make_history_range_url(self.pipe, start_time_inclusive, end_time_exclusive, request_id)
        logger.info('Get history range request url: %s', target_url)

        # Send out our request!
        headers = self._get_headers()
        response = self.restful.get(target_url, headers=headers)

        return response




# Copyright (c) 2012-2017 Snowflake Computing Inc. All rights reserved.

"""
tokentools.py - provides services for automatically creating and renewing
JWT tokens for authenticating to Snowflake
"""

from typing import Text
from datetime import timedelta, datetime
from logging import getLogger

import jwt

logger = getLogger(__name__)

ISSUER = "iss"
EXPIRE_TIME = "exp"
ISSUE_TIME = "iat"


class SecurityManager(object):
    """
    Given a private key, username, and account, signs and creates tokens for use in Snowflake Ingest Requests
    """
    LIFETIME = timedelta(minutes=59)  # The tokens will have a 59 minute lifetime
    RENEWAL_DELTA = timedelta(minutes=54)  # Tokens will be renewed after 54 minutes
    ALGORITHM = "RS256"  # Tokens will be generated using RSA with SHA256

    def __init__(self, account: Text, user: Text, private_key: Text,
                 lifetime: timedelta = LIFETIME, renewal_delay: timedelta = RENEWAL_DELTA):
        """
        __init__ creates a security manager with the specified context arguments
        :param account: the account in which data is being loaded
        :param user: The user who is loading these files
        :param private_key: the private key we'll use for signing tokens
        :param lifetime: how long this key will live (in minutes)
        :param renewal_delay: how long until the security manager should renew the key
        """

        logger.info(
            """Creating Security Manager with arguments
            account : %s, user : %s, lifetime : %s, renewal_delay : %s""",
            account, user, lifetime, renewal_delay)

        self.account = account.upper()  # Snowflake account names are canonically in all caps
        self.user = user.upper()  # Snowflake user names are also in all caps by default
        self.qualified_username = self.account + "." + self.user  # Generate the full user name
        self.lifetime = lifetime  # the timedelta until our tokens expire
        self.renewal_delay = renewal_delay  # the timedelta until we renew the token
        self.private_key = private_key  # stash the private key
        self.renew_time = datetime.utcnow()  # We need to renew the token NOW
        self.token = None  # We initially have no token

    def get_token(self) -> Text:
        """
        Regenerates the current token if and only if we have exceeded the renewal time
        bounds set
        :return: the new token
        """
        now = datetime.utcnow()  # Fetch the current time

        # If the token has expired, or doesn't exist, regenerate it
        if self.token is None or self.renew_time <= now:
            logger.info("Renewing token because renewal time (%s) is eclipsed by present time (%s)",
                        self.renew_time, now)
            # Calculate the next time we need to renew the token
            self.renew_time = now + self.renewal_delay

            # Create our payload
            payload = {

                # The issuer is the fully qualified user name
                ISSUER: self.qualified_username,

                # The payload was issued at this point it time
                ISSUE_TIME: now,

                # The token should no longer be accepted after our lifetime has elapsed
                EXPIRE_TIME: now + SecurityManager.LIFETIME
            }

            # Regenerate the actual token
            self.token = jwt.encode(payload, self.private_key, algorithm=SecurityManager.ALGORITHM)
            logger.info("New Token is %s", self.token)

        return self.token.decode('utf-8')



from typing import Text, Dict, Any

# import to use ocsp module in python connector
# this import will inject ocsp check method into
# botocore.vendored.requests library
import snowflake.connector

# use requsts library bundled in botocore
from botocore.vendored import requests
from botocore.vendored.requests import Response
import time
from ..error import IngestResponseError

from logging import getLogger
logger = getLogger(__name__)

# default timeout in seconds for a rest request
DEFAULT_REQUEST_TIMEOUT = 1 * 60


class SnowflakeRestful(object):
    """
        A simple wrapper over python request library to handle retry
    """
    def post(self, url: Text, json: Dict, headers: Dict) -> Dict[Text, Any]:
        """
        Http POST request
        :param url: request url,
        :param json: post request body
        :param headers: request headers, authentication etc
        :return: response payload
        """
        return self._exec_request_with_retry(url=url, method='POST', json=json, headers=headers)

    def get(self, url: Text, headers: Dict) -> Dict[Text, Any]:
        """
        Http GET request
        :param url:
        :param headers:
        :return:
        """
        return self._exec_request_with_retry(url=url, method='GET', headers=headers)

    def _exec_request_with_retry(self, url: Text, method: Text, headers: Dict = None, json: Dict = None) \
            -> Dict[Text, Any]:

        class RetryCtx(object):
            def __init__(self, timeout=None):
                self.retry_count = 0
                self.total_timeout = timeout
                self.next_sleep_time = 1
                self._request_start_time = time.time()

            def sleep_time(self):
                """
                :return: time in seconds to sleep next time, -1 if should not sleep
                """
                if time.time() - self._request_start_time > self.total_timeout:
                    logger.info('Request timeout reached.')
                    return -1
                else:
                    # exponential backoff
                    this_sleep_time = self.next_sleep_time
                    self.next_sleep_time = this_sleep_time * 2
                    self.retry_count += 1
                    logger.info('Retried request. Backoff time %d, Retry count %d',
                                this_sleep_time, self.retry_count)
                    return this_sleep_time

        retry_context = RetryCtx(DEFAULT_REQUEST_TIMEOUT)

        while True:
            response = self._exec_request(url=url, method=method, headers=headers, json=json)

            if response.ok:
                return response.json()
            elif self._can_retry(response.status_code):
                next_sleep_time = retry_context.sleep_time()
                if next_sleep_time > 0:
                    time.sleep(next_sleep_time)
                    continue

            raise IngestResponseError(response)

    def _exec_request(self, url: Text, method: Text, headers: Dict = None, json: Dict = None) -> Response:
        return requests.request(method=method,
                                url=url,
                                headers=headers,
                                json=json)

    @staticmethod
    def _can_retry(http_code):
        return http_code == 408 or 500 <= http_code < 600



# Copyright (c) 2012-2017 Snowflake Computing Inc. All rights reserved.

"""
request_builder.py - Builds the URIs and http requests that we send for ingest
requests and history requests
"""

from uuid import uuid4, UUID
from furl import furl
from typing import Text
from logging import getLogger

# Create a logger for this module
logger = getLogger(__name__)

# URI Construction constants we use for making a target URL
DEFAULT_HOST_FMT = "{}.snowflakecomputing.com"  # by default we target the Snowflake US instance
DEFAULT_PORT = 443  # we also target https by default implying 443
DEFAULT_SCHEME = "https"  # we also need to set the scheme to HTTPS

# Our format strings for the endpoints we need target
INGEST_ENDPOINT_FORMAT = "/v1/data/pipes/{0}/insertFiles"  # The template for an ingest request endpoint
HISTORY_ENDPOINT_FORMAT = "/v1/data/pipes/{0}/insertReport"  # The template for an ingest history endpoint
HISTORY_SCAN_ENDPOINT_FORMAT = "/v1/data/pipes/{0}/loadHistoryScan" # The template for ingest history range endpoint

# Parameter used to pass along request UUIDs
REQUEST_ID_PARAMETER = "requestId"
# Parameter that we use for setting the stage for this url
STAGE_PARAMETER = "stage"
# Parameters for insertReport endpoint
RECENT_HISTORY_IN_SECONDS_PARAMETER = 'recentSeconds'
HISTORY_BEGIN_MARK = 'beginMark'
# Parameters for loadHistoryScan endpoint
HISTORY_RANGE_START_INCLUSIVE = 'startTimeInclusive'
HISTORY_RANGE_END_EXCLUSIVE = 'endTimeExclusive'

# Method to generate an the URL for an ingest request for a given table, and stage
class URLGenerator(object):
    """
    URLGenerator - this class handles creating the URLs for a requests to
    the Snowflake service
    """
    def __init__(self, host: Text, scheme: Text = DEFAULT_SCHEME, port: int = DEFAULT_PORT):
        """
        This constructor simply stashes the basic portions of the request URL such that the user
        doesn't have to repeatedly provide them
        """
        self.scheme = scheme
        self.host = host
        self.port = port

    def _make_base_url(self, uuid: UUID = None) -> furl:
        """
        _makeBaseURL - generates the common base URL for all of our requests
        :param uuid: a UUID we want to attach to the
        :return: the furl wrapper around our unfinished base url
        """
        base = furl()  # Create an uninitialized base URI object
        base.host = self.host  # set the host name
        base.port = self.port  # set the port number
        base.scheme = self.scheme  # set the access scheme

        # if we have no uuid to attach to this request, generate one
        if uuid is None:
            uuid = uuid4()

        # Set the request id parameter uuid
        base.args[REQUEST_ID_PARAMETER] = str(uuid)

        return base

    def make_ingest_url(self, pipe: Text, uuid: UUID = None) -> Text:
        """
        make_ingest_url - creates a textual representation of the target url we need to hit for
        ingesting files
        :param pipe: the pipe which we want to use to ingest files (fully qualified)
        :param uuid: an optional UUID argument to tag this request
        :return: the completed URL
        """

        # Compute the base url
        builder = self._make_base_url(uuid)

        # Set the path for the ingest url
        builder.path = INGEST_ENDPOINT_FORMAT.format(pipe)

        return builder.url

    def make_history_url(self, pipe: Text, recent_seconds: int = None,
            begin_mark: Text = None, uuid: UUID = None) -> Text:
        """
        make_history_url - creates a textual representation of the target url we need to hit for
        history requests
        :param pipe: the pipe for which we want to see the see history
        :param uuid: an optional UUID argument to tag this request
        :param recent_seconds: an optional argument to specify recent seconds
        :param begin_mark: an optional argument used to indicate from which record should next reponse
                           return
        :return: the completed URL
        """

        # Compute the base url
        builder = self._make_base_url(uuid)

        # Set the path for the history url
        builder.path = HISTORY_ENDPOINT_FORMAT.format(pipe)
        
        if recent_seconds is not None:
            builder.args[RECENT_HISTORY_IN_SECONDS_PARAMETER] = str(recent_seconds)

        if begin_mark is not None:
            builder.args[HISTORY_BEGIN_MARK] = str(begin_mark)

        return builder.url

    def make_history_range_url(self, pipe: Text, start_time_inclusive: Text,
            end_time_exclusive: Text = None, uuid: UUID = None) -> Text:
        """
        make_history_url - creates [M#Ka textual representation of the target url we need to hit for
        history range scan requests
        :param pipe: the pipe for which we want to see the see history
        :param uuid: an optional UUID argument to tag this request
        :param start_time_inclusive: Timestamp in ISO-8601 format. Start of the time range to retrieve load history data.
        :param end_time_exclusive: Timestamp in ISO-8601 format. End of the time range to retrieve load history data.
                                    If omitted, then CURRENT_TIMESTAMP() is used as the end of the range.
        :return: the completed URL
        """

        # Compute the base url
        builder = self._make_base_url(uuid)

        # Set the path for the history url
        builder.path = HISTORY_SCAN_ENDPOINT_FORMAT.format(pipe)

        builder.args[HISTORY_RANGE_START_INCLUSIVE] = start_time_inclusive

        if end_time_exclusive is not None:
            builder.args[HISTORY_RANGE_END_EXCLUSIVE] = end_time_exclusive

        return builder.url



#! /usr/bin/env python3

# We need this to define our package
from setuptools import setup

# We use this to find and deploy our unittests
import unittest
import os
# We need to know the version to backfill some dependencies
from sys import version_info, exit
# Define our list of installation dependencies
DEPENDS = ["pyjwt", "snowflake-connector-python", "furl", "cryptography"]

# If we're at version less than 3.4 - fail
if version_info[0] < 3 or version_info[1] < 4:
    exit("Unsupported version of Python. Minimum version for the Ingest SDK is 3.4")

# If we're at version 3.4, backfill the typing library
elif version_info[1] == 4:
    DEPENDS.append("typing")

here = os.path.abspath(os.path.dirname(__file__))

def test_suite():
    """
    Defines the test suite for the snowflake ingest SDK
    """
    loader = unittest.TestLoader()
    return loader.discover("tests", pattern="test_*.py")

about = {}
with open(os.path.join(here, 'snowflake', 'ingest', 'version.py'),
          mode='r', encoding='utf-8') as f:
    exec(f.read(), about)

__version__ = about['__version__']

if 'SF_BUILD_NUMBER' in os.environ:
    __version__ += ('.' + str(os.environ['SF_BUILD_NUMBER']))

with open(os.path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
        long_description = f.read()

setup(
    name='snowflake_ingest',
    version=__version__,
    description='Official SnowflakeDB File Ingest SDK',
    long_description=long_description,
    author='Snowflake Computing',
    author_email='support@snowflake.net',
    url='https://www.snowflake.net',
    packages=['snowflake.ingest',
              'snowflake.ingest.utils'],
    license='Apache',
    keywords="snowflake ingest sdk copy loading",

    package_data={
        'snowflake.ingest':['*.rst', 'LICENSE']
    },
    # From here we describe the package classifiers
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Database"
    ],
    # Now we describe the dependencies
    install_requires=DEPENDS,
    # At last we set the test suite
    test_suite="setup.test_suite"
)
